@@ -37,6 +37,42 @@ def test_idempotent():
     assert fs3.storage == fs.storage
 
 
+def test_blockcache_workflow(ftp_writable, tmp_path):
+    host, port, user, pw = ftp_writable
+    fs = FTPFileSystem(host, port, user, pw)
+    with fs.open("/out", "wb") as f:
+        f.write(b"test\n" * 4096)
+
+    fs_kwargs = dict(
+        skip_instance_cache=True,
+        cache_storage=str(tmp_path),
+        target_protocol="ftp",
+        target_options={"host": host, "port": port, "username": user, "password": pw},
+    )
+
+    # Open the blockcache and read a little bit of the data
+    fs = fsspec.filesystem("blockcache", **fs_kwargs)
+    with fs.open("/out", "rb", block_size=5) as f:
+        assert f.read(5) == b"test\n"
+
+    # Save the cache/close it
+    fs.save_cache()
+    del fs
+
+    # Check that cache file only has the first two blocks
+    with open(tmp_path / "cache", "rb") as f:
+        cache = pickle.load(f)
+        assert "ftp:///out" in cache
+        assert cache["ftp:///out"]["blocks"] == [0, 1]
+
+    # Reopen the same cache and read some more...
+    fs = fsspec.filesystem("blockcache", **fs_kwargs)
+    with fs.open("/out", block_size=5) as f:
+        assert f.read(5) == b"test\n"
+        f.seek(30)
+        assert f.read(5) == b"test\n"
+
+
 @pytest.mark.parametrize("impl", ["filecache", "blockcache"])
 def test_workflow(ftp_writable, impl):
     host, port, user, pw = ftp_writable
@@ -79,6 +115,87 @@ def test_glob(ftp_writable, impl):
     assert fs.glob("/ou*") == ["/out", "/out2"]
 
 
+def test_write():
+    tmp = str(tempfile.mkdtemp())
+    fn = tmp + "afile"
+    url = "simplecache::file://" + fn
+    with fsspec.open(url, "wb") as f:
+        f.write(b"hello")
+        assert fn not in f.name
+        assert not os.listdir(tmp)
+
+    assert open(fn, "rb").read() == b"hello"
+
+
+def test_clear():
+    import tempfile
+
+    origin = tempfile.mkdtemp()
+    cache1 = tempfile.mkdtemp()
+    data = b"test data"
+    f1 = os.path.join(origin, "afile")
+    with open(f1, "wb") as f:
+        f.write(data)
+
+    # populates first cache
+    fs = fsspec.filesystem("filecache", target_protocol="file", cache_storage=cache1)
+    assert fs.cat(f1) == data
+
+    assert "cache" in os.listdir(cache1)
+    assert len(os.listdir(cache1)) == 2
+    assert fs._check_file(f1)
+
+    fs.clear_cache()
+    assert not fs._check_file(f1)[0]
+    assert len(os.listdir(cache1)) < 2
+
+
+def test_pop():
+    import tempfile
+
+    origin = tempfile.mkdtemp()
+    cache1 = tempfile.mkdtemp()
+    cache2 = tempfile.mkdtemp()
+    data = b"test data"
+    f1 = os.path.join(origin, "afile")
+    f2 = os.path.join(origin, "bfile")
+    with open(f1, "wb") as f:
+        f.write(data)
+    with open(f2, "wb") as f:
+        f.write(data)
+
+    # populates first cache
+    fs = fsspec.filesystem("filecache", target_protocol="file", cache_storage=cache1)
+    fs.cat(f1)
+
+    # populates last cache if file not found in first cache
+    fs = fsspec.filesystem(
+        "filecache", target_protocol="file", cache_storage=[cache1, cache2]
+    )
+    assert fs.cat(f2) == data
+    assert len(os.listdir(cache2)) == 2
+    assert fs._check_file(f1)
+    with pytest.raises(PermissionError):
+        fs.pop_from_cache(f1)
+    fs.pop_from_cache(f2)
+    assert len(os.listdir(cache2)) == 1
+    assert fs._check_file(f2)[0] is False
+    assert fs._check_file(f1)
+
+
+def test_write_pickle_context():
+    tmp = str(tempfile.mkdtemp())
+    fn = tmp + "afile"
+    url = "simplecache::file://" + fn
+    f = fsspec.open(url, "wb").open()
+    f.write(b"hello ")
+    f.flush()
+    with pickle.loads(pickle.dumps(f)) as f2:
+        f2.write(b"world")
+
+    assert open(fn, "rb").read() == b"hello world"
+
+
 def test_blocksize(ftp_writable):
     host, port, user, pw = ftp_writable
     fs = FTPFileSystem(host, port, user, pw)
@@ -320,14 +437,14 @@ def test_filecache_with_checks():
 
 
 @pytest.mark.parametrize("impl", ["filecache", "simplecache", "blockcache"])
-def test_takes_fs_instance(impl):
+@pytest.mark.parametrize("fs", ["local", "multi"], indirect=["fs"])
+def test_takes_fs_instance(impl, fs):
     origin = tempfile.mkdtemp()
     data = b"test data"
     f1 = os.path.join(origin, "afile")
     with open(f1, "wb") as f:
         f.write(data)
 
-    fs = fsspec.filesystem("file")
     fs2 = fsspec.filesystem(impl, fs=fs)
 
     assert fs2.cat(f1) == data
@@ -385,3 +502,23 @@ def test_with_compression(impl, compression):
         assert f.read() == data
         assert "data" in os.listdir(cachedir)
         assert open(os.path.join(cachedir, "data"), "rb").read() == data
+
+
+@pytest.mark.parametrize("protocol", ["simplecache", "filecache"])
+def test_again(protocol):
+    fn = "memory://afile"
+    with fsspec.open(fn, "wb") as f:
+        f.write(b"hello")
+    d2 = tempfile.mkdtemp()
+    lurl = fsspec.open_local(f"{protocol}::{fn}", **{protocol: {"cache_storage": d2}})
+    assert os.path.exists(lurl)
+    assert d2 in lurl
+    assert open(lurl, "rb").read() == b"hello"
+
+    # remove cache dir
+    shutil.rmtree(d2)
+    assert not os.path.exists(lurl)
+
+    # gets recreated
+    lurl = fsspec.open_local(f"{protocol}::{fn}", **{protocol: {"cache_storage": d2}})
+    assert open(lurl, "rb").read() == b"hello"
